# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from .. import cmdline, tests
from .features import backslashdir_feature


class TestSplitter(tests.TestCase):

    def assertAsTokens(self, expected, line, single_quotes_allowed=False):
        s = cmdline.Splitter(line, single_quotes_allowed=single_quotes_allowed)
        self.assertEqual(expected, list(s))

    def test_simple(self):
        self.assertAsTokens([(False, 'foo'), (False, 'bar'), (False, 'baz')],
                            'foo bar baz')

    def test_ignore_multiple_spaces(self):
        self.assertAsTokens([(False, 'foo'), (False, 'bar')], 'foo  bar')

    def test_ignore_leading_space(self):
        self.assertAsTokens([(False, 'foo'), (False, 'bar')], '  foo bar')

    def test_ignore_trailing_space(self):
        self.assertAsTokens([(False, 'foo'), (False, 'bar')], 'foo bar  ')

    def test_posix_quotations(self):
        self.assertAsTokens([(True, 'foo bar')], "'foo bar'",
                            single_quotes_allowed=True)
        self.assertAsTokens([(True, 'foo bar')], "'fo''o b''ar'",
                            single_quotes_allowed=True)
        self.assertAsTokens([(True, 'foo bar')], '"fo""o b""ar"',
                            single_quotes_allowed=True)
        self.assertAsTokens([(True, 'foo bar')], '"fo"\'o b\'"ar"',
                            single_quotes_allowed=True)

    def test_nested_quotations(self):
        self.assertAsTokens([(True, 'foo"" bar')], "\"foo\\\"\\\" bar\"")
        self.assertAsTokens([(True, 'foo\'\' bar')], "\"foo'' bar\"")
        self.assertAsTokens([(True, 'foo\'\' bar')], "\"foo'' bar\"",
                            single_quotes_allowed=True)
        self.assertAsTokens([(True, 'foo"" bar')], "'foo\"\" bar'",
                            single_quotes_allowed=True)

    def test_empty_result(self):
        self.assertAsTokens([], '')
        self.assertAsTokens([], '    ')

    def test_quoted_empty(self):
        self.assertAsTokens([(True, '')], '""')
        self.assertAsTokens([(False, "''")], "''")
        self.assertAsTokens([(True, '')], "''", single_quotes_allowed=True)
        self.assertAsTokens([(False, 'a'), (True, ''), (False, 'c')],
                            'a "" c')
        self.assertAsTokens([(False, 'a'), (True, ''), (False, 'c')],
                            "a '' c", single_quotes_allowed=True)

    def test_unicode_chars(self):
        self.assertAsTokens([(False, 'f\xb5\xee'), (False, '\u1234\u3456')],
                            'f\xb5\xee \u1234\u3456')

    def test_newline_in_quoted_section(self):
        self.assertAsTokens([(True, 'foo\nbar\nbaz\n')], '"foo\nbar\nbaz\n"')
        self.assertAsTokens([(True, 'foo\nbar\nbaz\n')], "'foo\nbar\nbaz\n'",
                            single_quotes_allowed=True)

    def test_escape_chars(self):
        self.assertAsTokens([(False, 'foo\\bar')], 'foo\\bar')

    def test_escape_quote(self):
        self.assertAsTokens([(True, 'foo"bar')], '"foo\\"bar"')
        self.assertAsTokens([(True, 'foo\\"bar')], '"foo\\\\\\"bar"')
        self.assertAsTokens([(True, 'foo\\bar')], '"foo\\\\"bar"')

    def test_double_escape(self):
        self.assertAsTokens([(True, 'foo\\\\bar')], '"foo\\\\bar"')
        self.assertAsTokens([(False, 'foo\\\\bar')], "foo\\\\bar")

    def test_multiple_quoted_args(self):
        self.assertAsTokens([(True, 'x x'), (True, 'y y')],
                            '"x x" "y y"')
        self.assertAsTokens([(True, 'x x'), (True, 'y y')],
                            '"x x" \'y y\'', single_quotes_allowed=True)

    def test_n_backslashes_handling(self):
        # https://bugs.launchpad.net/bzr/+bug/528944
        # actually we care about the doubled backslashes when they're
        # represents UNC paths.
        # But in fact there is too much weird corner cases
        # (see https://bugs.launchpad.net/tortoisebzr/+bug/569050)
        # so to reproduce every bit of windows command-line handling
        # could be not worth of efforts?
        self.requireFeature(backslashdir_feature)
        self.assertAsTokens([(True, r'\\host\path')], r'"\\host\path"')
        self.assertAsTokens([(False, r'\\host\path')], r'\\host\path')
        # handling of " after the 2n and 2n+1 backslashes
        # inside and outside the quoted string
        self.assertAsTokens([(True, r'\\'), (False, r'*.py')], r'"\\\\" *.py')
        self.assertAsTokens([(True, r'\\" *.py')], r'"\\\\\" *.py"')
        self.assertAsTokens([(True, r'\\ *.py')], r'\\\\" *.py"')
        self.assertAsTokens([(False, r'\\"'), (False, r'*.py')],
                            r'\\\\\" *.py')
        self.assertAsTokens([(True, '\\\\')], '"\\\\')



# Copyright (C) 2008-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test directory service implementation"""

from .. import transport, urlutils
from ..directory_service import (AliasDirectory, DirectoryServiceRegistry,
                                 InvalidLocationAlias, UnsetLocationAlias,
                                 directories)
from . import TestCase, TestCaseWithTransport


class FooService:
    """A directory service that maps the name to a FILE url"""

    # eg 'file:///foo' on Unix, or 'file:///C:/foo' on Windows
    base = urlutils.local_path_to_url('/foo')

    def look_up(self, name, url, purpose=None):
        return self.base + name


class TestDirectoryLookup(TestCase):

    def setUp(self):
        super().setUp()
        self.registry = DirectoryServiceRegistry()
        self.registry.register('foo:', FooService, 'Map foo URLs to http urls')

    def test_get_directory_service(self):
        directory, suffix = self.registry.get_prefix('foo:bar')
        self.assertIs(FooService, directory)
        self.assertEqual('bar', suffix)

    def test_dereference(self):
        self.assertEqual(FooService.base + 'bar',
                         self.registry.dereference('foo:bar'))
        self.assertEqual(FooService.base + 'bar',
                         self.registry.dereference('foo:bar', purpose='write'))
        self.assertEqual('baz:qux', self.registry.dereference('baz:qux'))
        self.assertEqual(
            'baz:qux',
            self.registry.dereference('baz:qux', purpose='write'))

    def test_get_transport(self):
        directories.register('foo:', FooService, 'Map foo URLs to http urls')
        self.addCleanup(directories.remove, 'foo:')
        self.assertEqual(FooService.base + 'bar/',
                         transport.get_transport('foo:bar').base)


class OldService:
    """A directory service that maps the name to a FILE url"""

    # eg 'file:///foo' on Unix, or 'file:///C:/foo' on Windows
    base = urlutils.local_path_to_url('/foo')

    def look_up(self, name, url):
        return self.base + name


class TestOldDirectoryLookup(TestCase):
    """Test compatibility with older implementations of Directory
    that don't support the purpose argument."""

    def setUp(self):
        super().setUp()
        self.registry = DirectoryServiceRegistry()
        self.registry.register('old:', OldService, 'Map foo URLs to http urls')

    def test_dereference(self):
        self.assertEqual(OldService.base + 'bar',
                         self.registry.dereference('old:bar'))
        self.assertEqual(OldService.base + 'bar',
                         self.registry.dereference('old:bar', purpose='write'))
        self.assertEqual('baz:qux', self.registry.dereference('baz:qux'))
        self.assertEqual(
            'baz:qux',
            self.registry.dereference('baz:qux', purpose='write'))


class TestAliasDirectory(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self.branch = self.make_branch('.')

    def assertAliasFromBranch(self, setter, value, alias):
        setter(value)
        self.assertEqual(value, directories.dereference(alias))

    def test_lookup_parent(self):
        self.assertAliasFromBranch(self.branch.set_parent, 'http://a',
                                   ':parent')

    def test_lookup_submit(self):
        self.assertAliasFromBranch(self.branch.set_submit_branch, 'http://b',
                                   ':submit')

    def test_lookup_public(self):
        self.assertAliasFromBranch(self.branch.set_public_branch, 'http://c',
                                   ':public')

    def test_lookup_bound(self):
        self.assertAliasFromBranch(self.branch.set_bound_location, 'http://d',
                                   ':bound')

    def test_lookup_push(self):
        self.assertAliasFromBranch(self.branch.set_push_location, 'http://e',
                                   ':push')

    def test_lookup_this(self):
        self.assertEqual(self.branch.base, directories.dereference(':this'))

    def test_extra_path(self):
        self.assertEqual(urlutils.join(self.branch.base, 'arg'),
                         directories.dereference(':this/arg'))

    def test_lookup_badname(self):
        e = self.assertRaises(InvalidLocationAlias,
                              directories.dereference, ':booga')
        self.assertEqual('":booga" is not a valid location alias.',
                         str(e))

    def test_lookup_badvalue(self):
        e = self.assertRaises(UnsetLocationAlias,
                              directories.dereference, ':parent')
        self.assertEqual('No parent location assigned.', str(e))

    def test_register_location_alias(self):
        self.addCleanup(AliasDirectory.branch_aliases.remove, "booga")
        AliasDirectory.branch_aliases.register("booga",
                                               lambda b: "UHH?", help="Nobody knows")
        self.assertEqual("UHH?", directories.dereference(":booga"))


class TestColocatedDirectory(TestCaseWithTransport):

    def test_lookup_non_default(self):
        default = self.make_branch('.')
        non_default = default.controldir.create_branch(name='nondefault')
        self.assertEqual(non_default.base,
                         directories.dereference('co:nondefault'))

    def test_lookup_default(self):
        default = self.make_branch('.')
        non_default = default.controldir.create_branch(name='nondefault')
        self.assertEqual(urlutils.join_segment_parameters(default.controldir.user_url,
                                                          {"branch": ""}), directories.dereference('co:'))

    def test_no_such_branch(self):
        # No error is raised in this case, that is up to the code that actually
        # opens the branch.
        default = self.make_branch('.')
        self.assertEqual(urlutils.join_segment_parameters(default.controldir.user_url,
                                                          {"branch": "foo"}), directories.dereference('co:foo'))



# Copyright (C) 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import time
from typing import Callable

from . import controldir, errors, osutils, registry, trace
from .branch import Branch
from .i18n import gettext
from .revision import NULL_REVISION

format_registry = registry.Registry[str, Callable]()


def send(target_branch, revision, public_branch, remember,
         format, no_bundle, no_patch, output, from_, mail_to, message, body,
         to_file, strict=None):
    possible_transports = []
    tree, branch = controldir.ControlDir.open_containing_tree_or_branch(
        from_, possible_transports=possible_transports)[:2]
    # we may need to write data into branch's repository to calculate
    # the data to send.
    with branch.lock_write():
        if output is None:
            config_stack = branch.get_config_stack()
            if mail_to is None:
                mail_to = config_stack.get('submit_to')
            mail_client = config_stack.get('mail_client')(config_stack)
            if (not getattr(mail_client, 'supports_body', False)
                    and body is not None):
                raise errors.CommandError(gettext(
                    'Mail client "%s" does not support specifying body') %
                    mail_client.__class__.__name__)
        if remember and target_branch is None:
            raise errors.CommandError(gettext(
                '--remember requires a branch to be specified.'))
        stored_target_branch = branch.get_submit_branch()
        remembered_target_branch = None
        if target_branch is None:
            target_branch = stored_target_branch
            remembered_target_branch = "submit"
        else:
            # Remembers if asked explicitly or no previous location is set
            if remember or (
                    remember is None and stored_target_branch is None):
                branch.set_submit_branch(target_branch)
        if target_branch is None:
            target_branch = branch.get_parent()
            remembered_target_branch = "parent"
        if target_branch is None:
            raise errors.CommandError(gettext('No submit branch known or'
                                              ' specified'))
        if remembered_target_branch is not None:
            trace.note(gettext('Using saved {0} location "{1}" to determine '
                               'what changes to submit.').format(
                remembered_target_branch,
                target_branch))

        submit_branch = Branch.open(target_branch,
                                    possible_transports=possible_transports)
        possible_transports.append(submit_branch.controldir.root_transport)
        if mail_to is None or format is None:
            if mail_to is None:
                mail_to = submit_branch.get_config_stack().get(
                    'child_submit_to')
            if format is None:
                formatname = submit_branch.get_child_submit_format()
                try:
                    format = format_registry.get(formatname)
                except KeyError:
                    raise errors.CommandError(
                        gettext("No such send format '%s'.") % formatname)

        stored_public_branch = branch.get_public_branch()
        if public_branch is None:
            public_branch = stored_public_branch
        # Remembers if asked explicitly or no previous location is set
        elif (remember
              or (remember is None and stored_public_branch is None)):
            branch.set_public_branch(public_branch)
        if no_bundle and public_branch is None:
            raise errors.CommandError(gettext('No public branch specified or'
                                              ' known'))
        base_revision_id = None
        revision_id = None
        if revision is not None:
            if len(revision) > 2:
                raise errors.CommandError(gettext('bzr send takes '
                                                  'at most two one revision identifiers'))
            revision_id = revision[-1].as_revision_id(branch)
            if len(revision) == 2:
                base_revision_id = revision[0].as_revision_id(branch)
        if revision_id is None:
            if tree is not None:
                tree.check_changed_or_out_of_date(
                    strict, 'send_strict',
                    more_error='Use --no-strict to force the send.',
                    more_warning='Uncommitted changes will not be sent.')
            revision_id = branch.last_revision()
        if revision_id == NULL_REVISION:
            raise errors.CommandError(gettext('No revisions to submit.'))
        if format is None:
            format = format_registry.get()
        directive = format(branch, revision_id, target_branch,
                           public_branch, no_patch, no_bundle, message, base_revision_id,
                           submit_branch)
        if output is None:
            directive.compose_merge_request(mail_client, mail_to, body,
                                            branch, tree)
        else:
            if directive.multiple_output_files:
                if output == '-':
                    raise errors.CommandError(gettext('- not supported for '
                                                      'merge directives that use more than one output file.'))
                if not os.path.exists(output):
                    os.mkdir(output, 0o755)
                for (filename, lines) in directive.to_files():
                    path = os.path.join(output, filename)
                    with open(path, 'wb') as outfile:
                        outfile.writelines(lines)
            else:
                if output == '-':
                    outfile = to_file
                else:
                    outfile = open(output, 'wb')
                try:
                    outfile.writelines(directive.to_lines())
                finally:
                    if outfile is not to_file:
                        outfile.close()


def _send_4(branch, revision_id, target_branch, public_branch,
            no_patch, no_bundle, message,
            base_revision_id, local_target_branch=None):
    from breezy import merge_directive
    return merge_directive.MergeDirective2.from_objects(
        branch.repository, revision_id, time.time(),
        osutils.local_time_offset(), target_branch,
        public_branch=public_branch,
        include_patch=not no_patch,
        include_bundle=not no_bundle, message=message,
        base_revision_id=base_revision_id,
        local_target_branch=local_target_branch)


def _send_0_9(branch, revision_id, submit_branch, public_branch,
              no_patch, no_bundle, message,
              base_revision_id, local_target_branch=None):
    if not no_bundle:
        if not no_patch:
            patch_type = 'bundle'
        else:
            raise errors.CommandError(gettext('Format 0.9 does not'
                                              ' permit bundle with no patch'))
    else:
        if not no_patch:
            patch_type = 'diff'
        else:
            patch_type = None
    from breezy import merge_directive
    return merge_directive.MergeDirective.from_objects(
        branch.repository, revision_id, time.time(),
        osutils.local_time_offset(), submit_branch,
        public_branch=public_branch, patch_type=patch_type,
        message=message, local_target_branch=local_target_branch)


format_registry.register('4',
                         _send_4, 'Bundle format 4, Merge Directive 2 (default)')
format_registry.register('0.9',
                         _send_0_9, 'Bundle format 0.9, Merge Directive 1')
format_registry.default_key = '4'



# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Annotate."""

from dulwich.object_store import tree_lookup_path

from .. import osutils
from ..bzr.versionedfile import UnavailableRepresentation
from ..errors import NoSuchRevision
from ..graph import Graph
from ..revision import NULL_REVISION
from .mapping import decode_git_path, encode_git_path


class GitBlobContentFactory:
    """Static data content factory.

    This takes a fulltext when created and just returns that during
    get_bytes_as('fulltext').

    :ivar sha1: None, or the sha1 of the content fulltext.
    :ivar storage_kind: The native storage kind of this factory. Always
        'fulltext'.
    :ivar key: The key of this content. Each key is a tuple with a single
        string in it.
    :ivar parents: A tuple of parent keys for self.key. If the object has
        no parent information, None (as opposed to () for an empty list of
        parents).
     """

    def __init__(self, store, path, revision, blob_id):
        """Create a ContentFactory."""
        self.store = store
        self.key = (path, revision)
        self.storage_kind = 'git-blob'
        self.parents = None
        self.blob_id = blob_id
        self.size = None

    def get_bytes_as(self, storage_kind):
        if storage_kind == 'fulltext':
            return self.store[self.blob_id].as_raw_string()
        elif storage_kind == 'lines':
            return list(osutils.chunks_to_lines(self.store[self.blob_id].as_raw_chunks()))
        elif storage_kind == 'chunked':
            return self.store[self.blob_id].as_raw_chunks()
        raise UnavailableRepresentation(self.key, storage_kind,
                                        self.storage_kind)

    def iter_bytes_as(self, storage_kind):
        if storage_kind == 'lines':
            return iter(osutils.chunks_to_lines(self.store[self.blob_id].as_raw_chunks()))
        elif storage_kind == 'chunked':
            return iter(self.store[self.blob_id].as_raw_chunks())
        raise UnavailableRepresentation(self.key, storage_kind,
                                        self.storage_kind)


class GitAbsentContentFactory:
    """Absent data content factory.

    :ivar sha1: None, or the sha1 of the content fulltext.
    :ivar storage_kind: The native storage kind of this factory. Always
        'fulltext'.
    :ivar key: The key of this content. Each key is a tuple with a single
        string in it.
    :ivar parents: A tuple of parent keys for self.key. If the object has
        no parent information, None (as opposed to () for an empty list of
        parents).
     """

    def __init__(self, store, path, revision):
        """Create a ContentFactory."""
        self.store = store
        self.key = (path, revision)
        self.storage_kind = 'absent'
        self.parents = None
        self.size = None

    def get_bytes_as(self, storage_kind):
        raise ValueError

    def iter_bytes_as(self, storage_kind):
        raise ValueError


class AnnotateProvider:

    def __init__(self, change_scanner):
        self.change_scanner = change_scanner
        self.store = self.change_scanner.repository._git.object_store

    def _get_parents(self, path, text_revision):
        commit_id, mapping = (
            self.change_scanner.repository.lookup_bzr_revision_id(
                text_revision))
        text_parents = []
        path = encode_git_path(path)
        for commit_parent in self.store[commit_id].parents:
            try:
                (store, path, text_parent) = (
                    self.change_scanner.find_last_change_revision(
                        path, commit_parent))
            except KeyError:
                continue
            if text_parent not in text_parents:
                text_parents.append(text_parent)
        return tuple([
            (decode_git_path(path),
                self.change_scanner.repository.lookup_foreign_revision_id(p))
            for p in text_parents])

    def get_parent_map(self, keys):
        ret = {}
        for key in keys:
            (path, text_revision) = key
            if text_revision == NULL_REVISION:
                ret[key] = ()
                continue
            try:
                ret[key] = self._get_parents(path, text_revision)
            except KeyError:
                pass
        return ret

    def get_record_stream(self, keys, ordering, include_delta_closure):
        if ordering == 'topological':
            graph = Graph(self)
            keys = graph.iter_topo_order(keys)
        store = self.change_scanner.repository._git.object_store
        for (path, text_revision) in keys:
            try:
                commit_id, mapping = (
                    self.change_scanner.repository.lookup_bzr_revision_id(
                        text_revision))
            except NoSuchRevision:
                yield GitAbsentContentFactory(store, path, text_revision)
                continue

            try:
                tree_id = store[commit_id].tree
            except KeyError:
                yield GitAbsentContentFactory(store, path, text_revision)
                continue
            try:
                (mode, blob_sha) = tree_lookup_path(
                    store.__getitem__, tree_id, encode_git_path(path))
            except KeyError:
                yield GitAbsentContentFactory(store, path, text_revision)
            else:
                yield GitBlobContentFactory(
                    store, path, text_revision, blob_sha)



# Copyright (C) 2005, 2006 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# XXX: Some consideration of the problems that might occur if there are
# files whose id differs only in case.  That should probably be forbidden.


import os

from .... import errors, osutils
from .... import transport as _mod_transport
from .... import ui
from ....trace import mutter
from . import TransportStore


class VersionedFileStore(TransportStore):
    """Collection of many versioned files in a transport."""

    # TODO: Rather than passing versionedfile_kwargs, perhaps pass in a
    # transport factory callable?
    def __init__(self, transport, prefixed=False, precious=False,
                 dir_mode=None, file_mode=None,
                 versionedfile_class=None,
                 versionedfile_kwargs={},
                 escaped=False):
        super().__init__(transport,
                                                 dir_mode=dir_mode, file_mode=file_mode,
                                                 prefixed=prefixed, compressed=False, escaped=escaped)
        self._precious = precious
        self._versionedfile_class = versionedfile_class
        self._versionedfile_kwargs = versionedfile_kwargs
        # Used for passing get_scope to versioned file constructors;
        self.get_scope = None

    def filename(self, file_id):
        """Return the path relative to the transport root."""
        return self._relpath(file_id)

    def __iter__(self):
        suffixes = self._versionedfile_class.get_suffixes()
        ids = set()
        for relpath in self._iter_files_recursive():
            for suffix in suffixes:
                if relpath.endswith(suffix):
                    # TODO: use standard remove_suffix function
                    escaped_id = os.path.basename(relpath[:-len(suffix)])
                    file_id = self._mapper.unmap(escaped_id)[0]
                    if file_id not in ids:
                        ids.add(file_id)
                        yield file_id
                    break  # only one suffix can match

    def has_id(self, file_id):
        suffixes = self._versionedfile_class.get_suffixes()
        filename = self.filename(file_id)
        for suffix in suffixes:
            if not self._transport.has(filename + suffix):
                return False
        return True

    def get_empty(self, file_id, transaction):
        """Get an empty weave, which implies deleting the existing one first."""
        if self.has_id(file_id):
            self.delete(file_id, transaction)
        return self.get_weave_or_empty(file_id, transaction)

    def delete(self, file_id, transaction):
        """Remove file_id from the store."""
        suffixes = self._versionedfile_class.get_suffixes()
        filename = self.filename(file_id)
        for suffix in suffixes:
            self._transport.delete(filename + suffix)

    def _get(self, file_id):
        return self._transport.get(self.filename(file_id))

    def _put(self, file_id, f):
        fn = self.filename(file_id)
        try:
            return self._transport.put_file(fn, f, mode=self._file_mode)
        except _mod_transport.NoSuchFile:
            if not self._prefixed:
                raise
            self._transport.mkdir(os.path.dirname(fn), mode=self._dir_mode)
            return self._transport.put_file(fn, f, mode=self._file_mode)

    def get_weave(self, file_id, transaction, _filename=None):
        """Return the VersionedFile for file_id.

        :param _filename: filename that would be returned from self.filename for
        file_id. This is used to reduce duplicate filename calculations when
        using 'get_weave_or_empty'. FOR INTERNAL USE ONLY.
        """
        if _filename is None:
            _filename = self.filename(file_id)
        if transaction.writeable():
            w = self._versionedfile_class(_filename, self._transport, self._file_mode,
                                          get_scope=self.get_scope, **self._versionedfile_kwargs)
        else:
            w = self._versionedfile_class(_filename,
                                          self._transport,
                                          self._file_mode,
                                          create=False,
                                          access_mode='r',
                                          get_scope=self.get_scope,
                                          **self._versionedfile_kwargs)
        return w

    def _make_new_versionedfile(self, file_id, transaction,
                                known_missing=False, _filename=None):
        """Make a new versioned file.

        :param _filename: filename that would be returned from self.filename for
        file_id. This is used to reduce duplicate filename calculations when
        using 'get_weave_or_empty'. FOR INTERNAL USE ONLY.
        """
        if not known_missing and self.has_id(file_id):
            self.delete(file_id, transaction)
        if _filename is None:
            _filename = self.filename(file_id)
        try:
            # we try without making the directory first because thats optimising
            # for the common case.
            weave = self._versionedfile_class(_filename, self._transport, self._file_mode, create=True,
                                              get_scope=self.get_scope, **self._versionedfile_kwargs)
        except _mod_transport.NoSuchFile:
            if not self._prefixed:
                # unexpected error - NoSuchFile is expected to be raised on a
                # missing dir only and that only occurs when we are prefixed.
                raise
            dirname = osutils.dirname(_filename)
            self._transport.mkdir(dirname, mode=self._dir_mode)
            weave = self._versionedfile_class(_filename, self._transport,
                                              self._file_mode, create=True,
                                              get_scope=self.get_scope,
                                              **self._versionedfile_kwargs)
        return weave

    def get_weave_or_empty(self, file_id, transaction):
        """Return a weave, or an empty one if it doesn't exist."""
        # This is typically used from 'commit' and 'fetch/push/pull' where
        # we scan across many versioned files once. As such the small overhead
        # of calculating the filename before doing a cache lookup is more than
        # compensated for by not calculating the filename when making new
        # versioned files.
        _filename = self.filename(file_id)
        try:
            return self.get_weave(file_id, transaction, _filename=_filename)
        except _mod_transport.NoSuchFile:
            weave = self._make_new_versionedfile(file_id, transaction,
                                                 known_missing=True, _filename=_filename)
            return weave

    def _put_weave(self, file_id, weave, transaction):
        """Preserved here for upgrades-to-weaves to use."""
        myweave = self._make_new_versionedfile(file_id, transaction)
        myweave.insert_record_stream(weave.get_record_stream(
            [(version,) for version in weave.versions()],
            'topological', False))

    def total_size(self):
        count, bytes = super().total_size()
        return (count / len(self._versionedfile_class.get_suffixes())), bytes



# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for how merge directives interact with various repository formats.

Bundles contain the serialized form, so changes in serialization based on
repository effects the final bundle.
"""

from breezy import merge_directive
from breezy.bzr import chk_map
from breezy.bzr.tests.per_repository_vf import (
    TestCaseWithRepository, all_repository_vf_format_scenarios)
from breezy.tests.scenarios import load_tests_apply_scenarios

load_tests = load_tests_apply_scenarios


class TestMergeDirective(TestCaseWithRepository):

    scenarios = all_repository_vf_format_scenarios()

    def make_two_branches(self):
        builder = self.make_branch_builder('source')
        builder.start_series()
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', None)),
            ('add', ('f', b'f-id', 'file', b'initial content\n')),
            ], revision_id=b'A')
        builder.build_snapshot([b'A'], [
            ('modify', ('f', b'new content\n')),
            ], revision_id=b'B')
        builder.finish_series()
        b1 = builder.get_branch()
        b2 = b1.controldir.sprout('target', revision_id=b'A').open_branch()
        return b1, b2

    def create_merge_directive(self, source_branch, submit_url):
        return merge_directive.MergeDirective2.from_objects(
            source_branch.repository,
            source_branch.last_revision(),
            time=1247775710, timezone=0,
            target_branch=submit_url)

    def test_create_merge_directive(self):
        source_branch, target_branch = self.make_two_branches()
        directive = self.create_merge_directive(source_branch,
                                                target_branch.base)
        self.assertIsInstance(directive, merge_directive.MergeDirective2)

    def test_create_and_install_directive(self):
        source_branch, target_branch = self.make_two_branches()
        directive = self.create_merge_directive(source_branch,
                                                target_branch.base)
        chk_map.clear_cache()
        directive.install_revisions(target_branch.repository)
        rt = target_branch.repository.revision_tree(b'B')
        with rt.lock_read():
            self.assertEqualDiff(b'new content\n', rt.get_file_text('f'))
